/**
 * @file   dfa_decode.c
 * @author Akinobu LEE
 * @date   Mon Mar  7 15:31:00 2005
 * 
 * <JA>
 * @brief  åǥǥ(裲ѥ)ˤ뵭ʸˡ˴Ťñͽ¬
 *
 * Ϳ줿ФơDFA ʸˡ³ǽʼñνꤹ롥
 * ºݤˤ, Ÿͽ¬üե졼դñȥꥹ
 * ˻ĤäƤñΤߤŸ롥
 *
 * Julian Ǥϥ硼ȥݡñ٥ǵҤΥ硼ȥݡñ
 * и֤ʸˡǻꤹ롥ºݤϤǤϤꤷ֤
 * ɬݡʤᡤñͽ¬ˤƤϥ硼ȥݡ
 * ȶޤʤθɬפ롥Τᡤ
 * ñ콸˥硼ȥݡϡˤμñ콸ޤǸ
 * ñ콸˴ޤ롥ºݤˤ˥硼ȥݡ뤫ɤϡ
 * search_bestfirst_main.c ξԤΥӤȽǤ롥
 *
 * Julian Ǥ dfa_firstwords(), dfa_nextwords(), dfa_acceptable(),
 * dfa_eosscore() 2ѥΥᥤؿ wchmm_fbs() Ѥ롥
 * ʤ Julius Ǥ ngram_decode.c δؿȤ롥
 * </JA>
 * 
 * <EN>
 * @brief  Prediction of next words based on DFA grammar for stack decoding (2nd pass)
 *
 * Given a part-of-sentence hypothesis, these function determine a set of next
 * words allowed to be connected by the grammar.  Actually, only words in the
 * word trellis, which exist around the estimated word-end frame will be
 * expanded.
 *
 * In Julian, the location where a (short) pause can be inserted should be
 * explicitly specified by grammar, just like other words.  Since user does not
 * always place pause at the specified place, the decoder have to consider
 * the skipping of such short pause word for the next word prediction.
 *
 * If a short pause word is contained in the set of next word candidates,
 * word set next to the short pause word is further included in the word
 * candidates.  Whether short pause was actually inserted or not in the user
 * input will be determined by score in search_bestfirst_main.c.
 *
 * In Julian mode, dfa_firstwords(), dfa_nextwords(), dfa_acceptable()
 * and dfa_eosscore() are called from main search function wchmm_fbs().
 * When Julius mode, on the other hand, the corresponding functions in
 * ngram_decode.c will be used instead.
 * </EN>
 * 
 * $Revision:$
 * 
 */
/*
 * Copyright (c) 1991-2005 Kawahara Lab., Kyoto University
 * Copyright (c) 2000-2005 Shikano Lab., Nara Institute of Science and Technology
 * Copyright (c) 2005      Julius project team, Nagoya Institute of Technology
 * All rights reserved
 */

#include <julius.h>
#ifdef USE_DFA

/** 
 * <JA>
 * ʸˡˤäơʸƬ³ñǽͽ¬ñ췲Ȥ֤
 * 
 * @param nw [out] ñ콸γǼؤΥݥ
 * @param peseqlen [in] ϥե졼Ĺ
 * @param maxnw [in] @a nw εĹ
 * @param dfa [in] DFAʸˡ¤
 * 
 * @return ͽ¬줿ñ (顼 -1 ֤)
 * </JA>
 * <EN>
 * Return initial word set from grammar.
 * 
 * @param nw [out] pointer to hold the resulting next word set
 * @param peseqlen [in] input frame length
 * @param maxnw [in] maximum number of words that can be set in @a nw
 * @param dfa [in] DFA grammar structure
 * 
 * @return the number of predicted words, or -1 on error.
 * </EN>
 */
int
dfa_firstwords(NEXTWORD **nw, int peseqlen, int maxnw, DFA_INFO *dfa)
{
  DFA_ARC *arc;
  MULTIGRAM *m;
  int s, sb, se;
  int cate, iw, ns;
  int num = 0;

  for (m = gramlist; m; m = m->next) {
    if (m->active) {
      sb = m->state_begin;
      se = sb + m->dfa->state_num;
      for(s=sb;s<se;s++) {
	if ((dfa->st[s].status & INITIAL_S) != 0) { /* from initial state */
	  for (arc = dfa->st[s].arc; arc; arc = arc->next) {	/* for all arc */
	    cate = arc->label;	/* category ID */
	    ns = arc->to_state;	/* next DFA state ID */
	    /* all words within the category is expanded */
	    for (iw=0;iw<dfa->term.wnum[cate];iw++) {
	      nw[num]->id = dfa->term.tw[cate][iw]; /* word ID */
	      nw[num]->next_state = ns; /* next state */
	      nw[num]->can_insert_sp = FALSE; /* short pause should not inserted before this word */
	      num++;
	      if (num >= maxnw) return -1; /* buffer overflow */
	    }
	  }
	}
      }
    }
  }

  return num;
}

/** 
 * <JA>
 * ʬʸФơʸˡ˽äƼ³ñ췲֤
 * 
 * @param hypo [in] Ÿʬʸ
 * @param nw [out] ñ콸γǼؤΥݥ
 * @param maxnw [in] @a nw εĹ
 * @param dfa [in] DFAʸˡ¤
 * 
 * @return ͽ¬줿ñ (顼 -1 ֤)
 * </JA>
 * <EN>
 * Given a part-of-sentence hypothesis, returns the next word set defined
 * by DFA grammar.
 * 
 * @param hypo [in] the source part-of-sentene hypothesis
 * @param nw [out] pointer to hold the resulting next word set
 * @param maxnw [in] maximum number of words that can be set in @a nw
 * @param dfa [in] DFA grammar structure
 * 
 * @return the number of predicted words, or -1 on error.
 * </EN>
 */
int
dfa_nextwords(NODE *hypo, NEXTWORD **nw, int maxnw, DFA_INFO *dfa)
{
  DFA_ARC *arc, *arc2;
  int iw,cate,ns,cate2,ns2;
  int num = 0;

  /* hypo->state: current DFA state ID */
  for (arc = dfa->st[hypo->state].arc; arc; arc = arc->next) {/* for all arc */
    cate = arc->label;
    ns = arc->to_state;
    if (dfa->is_sp[cate]) {	/* short pause */
      /* expand one more next (not expand the short pause word itself) */
      for (arc2 = dfa->st[ns].arc; arc2; arc2 = arc2->next) {
	cate2 = arc2->label;
	ns2 = arc2->to_state;
	for (iw=0;iw<dfa->term.wnum[cate2];iw++) {
	  nw[num]->id = dfa->term.tw[cate2][iw];
	  nw[num]->next_state = ns2;
	  nw[num]->can_insert_sp = TRUE;
	  num++;
	  if (num >= maxnw) return -1; /* buffer overflow */
	}
      }
    } else {			/* not short pause */
      /* all words within the category is expanded */
      for (iw=0;iw<dfa->term.wnum[cate];iw++) {
	nw[num]->id = dfa->term.tw[cate][iw];
	nw[num]->next_state = ns;
	nw[num]->can_insert_sp = FALSE;
	num++;
	if (num >= maxnw) return -1; /* buffer overflow */
      }
    }
  }
  return num;
}


/** 
 * <JA>
 * ʬʸ⤬ʸˡʸȤƺǽ(ǽ)֤ˤ뤫ɤ֤
 * 
 * @param hypo [in] ʬʸ
 * @param dfa [in] DFAʸˡ¤
 * 
 * @return ǽ֤ˤȤ TRUE ԲǽʤȤ FALSE
 * </JA>
 * <EN>
 * Return whether the hypothesis is currently on final state
 * 
 * @param hypo [in] sentence hypothesis
 * @param dfa [in] DFA grammar structure
 * 
 * @return TRUE when on final state, or FALSE if not acceptable.
 * </EN>
 */
boolean
dfa_acceptable(NODE *hypo, DFA_INFO *dfa)
{
  if (dfa->st[hypo->state].status & ACCEPT_S) {
    return TRUE;
  } else {
    return FALSE;
  }
}

/* patch by kashima */
/** 
 * <JA>
 * ñ䤬οꤵ줿³ͽ¬ñȥꥹ
 * 뤫ɤå⤷ФΥȥꥹñؤΥݥ󥿤򥻥å
 * 롥ʤ³ϤȤǷޤΤǡǤϺŬʥȥꥹñ
 * ǤʤƤ褤
 * 
 * @param nword [i/o] ñ (бȥꥹñؤΥݥ󥿤
 * åȤ)
 * @param hypo [in] Ÿ
 * @param bt [in] ñȥꥹ¤
 * 
 * @return ñȥꥹͽ¬ն˼ñ줬¸ߤ TRUE¸
 * ʤ FALSE ֤
 * </JA>
 * <EN>
 * Check if the given nextword exists in the word trellis around the
 * estimated connection time.  If exist, set the pointer to the corresponding
 * trellis word to the nextword.  Since the best connection time will be
 * re-computed later, it need not to be an optimal one.
 * 
 * @param nword [i/o] next word candidate (pointer to the found trellis word
 * will be set)
 * @param hypo [in] source part-of-sentence hypothesis
 * @param bt [in] word trellis structure
 * 
 * @return TRUE if the nextword exists on the word trellis around the estimated
 * connection point, or FALSE if not exist.
 * </EN>
 */
boolean
dfa_look_around(NEXTWORD *nword, NODE *hypo, BACKTRELLIS *bt)
{
  int t,tm;
  int i;
  WORD_ID w;
  
  tm = hypo->estimated_next_t;	/* estimated connection time */

  /* look aound [tm-lookup_range..tm+lookup_range] frame */
  /* near the center is better:
     1. the first half (backward)   2. the second half (forward) */
  /* 1. backward */
  for(t = tm; t >= tm - lookup_range; t--) {
    if (t < 0) break;
     for (i=0;i<bt->num[t];i++) {
       w = (bt->rw[t][i])->wid;
       if(w == nword->id){	/* found */
         nword->tre = bt->rw[t][i];
         return TRUE;
       }
     }
  }
  /* 2. forward */
  for(t = tm + 1; t < tm + lookup_range; t++) {
    if (t > bt->framelen - 1) break;
    if (t >= hypo->bestt) break;
    for (i=0;i<bt->num[t];i++) {
      w = (bt->rw[t][i])->wid;
      if(w == nword->id){	/* found */
        nword->tre = bt->rw[t][i];
        return TRUE;
      }
    }
  }

  return FALSE;			/* not found */
}

#endif /* USE_DFA */
